@@ -1,10 +1,59 @@
-from abc import abstractmethod, ABC
+"""
+Blocks are a series of components that can be combined to create rich and
+interactive messages.
+See: https://api.slack.com/reference/block-kit/blocks?ref=bk
+"""
+from abc import ABC, abstractmethod
 from enum import Enum
 from json import dumps
 from typing import Any, Dict, List, Optional, Union
 from uuid import uuid4
-from .elements import Element, ElementType, Text, TextType
-from .errors import InvalidUsageError
+
+from slackblocks.elements import (
+    ChannelMultiSelectMenu,
+    ChannelSelectMenu,
+    CheckboxGroup,
+    ConversationMultiSelectMenu,
+    ConversationSelectMenu,
+    DatePicker,
+    Element,
+    ElementType,
+    ExternalMultiSelectMenu,
+    ExternalSelectMenu,
+    PlainTextInput,
+    RadioButtonGroup,
+    StaticMultiSelectMenu,
+    StaticSelectMenu,
+    UserMultiSelectMenu,
+    UserSelectMenu,
+)
+from slackblocks.errors import InvalidUsageError
+from slackblocks.objects import (
+    CompositionObject,
+    CompositionObjectType,
+    DispatchActionConfiguration,
+    Text,
+    TextLike,
+    TextType,
+)
+from slackblocks.utils import coerce_to_list
+
+ALLOWED_INPUT_ELEMENTS = (
+    PlainTextInput,
+    CheckboxGroup,
+    RadioButtonGroup,
+    DatePicker,
+    ChannelSelectMenu,
+    ChannelMultiSelectMenu,
+    ConversationSelectMenu,
+    ConversationMultiSelectMenu,
+    ExternalSelectMenu,
+    ExternalMultiSelectMenu,
+    StaticSelectMenu,
+    StaticMultiSelectMenu,
+    UserSelectMenu,
+    UserMultiSelectMenu,
+)
 
 
 class BlockType(Enum):
@@ -16,6 +65,7 @@ class BlockType(Enum):
     SECTION = "section"
     DIVIDER = "divider"
     IMAGE = "image"
+    INPUT = "input"
     ACTIONS = "actions"
     CONTEXT = "context"
     FILE = "file"
@@ -46,40 +96,56 @@ class Block(ABC):
         return dumps(self._resolve(), indent=4)
 
 
-class SectionBlock(Block):
+class ActionsBlock(Block):
     """
-    A section is one of the most flexible blocks available -
-    it can be used as a simple text block, in combination with text fields,
-    or side-by-side with any of the available block elements.
+    A block that is used to hold interactive elements.
     """
 
     def __init__(
         self,
-        text: Optional[Union[str, Text]] = None,
+        elements: Optional[List[Union[Element, CompositionObject]]] = None,
         block_id: Optional[str] = None,
-        fields: Optional[List[Text]] = None,
-        accessory: Optional[Element] = None,
     ):
-        super().__init__(type_=BlockType.SECTION, block_id=block_id)
-        if text:
-            if type(text) is Text:
-                self.text = text
-            else:
-                self.text = Text(text)
-        else:
-            self.text = text
-        self.fields = fields
-        self.accessory = accessory
+        super().__init__(type_=BlockType.ACTIONS, block_id=block_id)
+        self.elements = coerce_to_list(
+            elements, (Element, CompositionObject), allow_none=True, max_size=25
+        )
 
-    def _resolve(self) -> Dict[str, Any]:
-        section = self._attributes()
-        if self.text:
-            section["text"] = self.text._resolve()
-        if self.fields:
-            section["fields"] = [field._resolve() for field in self.fields]
-        if self.accessory:
-            section["accessory"] = self.accessory._resolve()
-        return section
+    def _resolve(self):
+        actions = self._attributes()
+        actions["elements"] = [element._resolve() for element in self.elements]
+        return actions
+
+
+class ContextBlock(Block):
+    """
+    Displays message context, which can include both images and text.
+    """
+
+    def __init__(
+        self,
+        elements: Optional[List[Union[Element, CompositionObjectType]]] = None,
+        block_id: Optional[str] = None,
+    ):
+        super().__init__(type_=BlockType.CONTEXT, block_id=block_id)
+        self.elements = []
+        for element in elements:
+            if (
+                element.type == CompositionObjectType.TEXT
+                or element.type == ElementType.IMAGE
+            ):
+                self.elements.append(element)
+            else:
+                raise InvalidUsageError(
+                    f"Context blocks can only hold image and text elements, not {element.type}"
+                )
+        if len(self.elements) > 10:
+            raise InvalidUsageError("Context blocks can hold a maximum of ten elements")
+
+    def _resolve(self) -> Dict[str, any]:
+        context = self._attributes()
+        context["elements"] = [element._resolve() for element in self.elements]
+        return context
 
 
 class DividerBlock(Block):
@@ -95,6 +161,41 @@ class DividerBlock(Block):
         return self._attributes()
 
 
+class FileBlock(Block):
+    """
+    Displays a remote file.
+    """
+
+    def __init__(self, external_id: str, source: str, block_id: Optional[str]):
+        super().__init__(type_=BlockType.FILE, block_id=block_id)
+        self.external_id = external_id
+        self.source = source
+
+    def _resolve(self) -> Dict[str, any]:
+        file = self._attributes()
+        file["external_id"] = self.external_id
+        file["source"] = self.source
+        return file
+
+
+class HeaderBlock(Block):
+    """
+    A header is a plain-text block that displays in a larger, bold font.
+    """
+
+    def __init__(self, text: Union[str, Text], block_id: Optional[str] = None):
+        super().__init__(type_=BlockType.HEADER, block_id=block_id)
+        if type(text) is Text:
+            self.text = text
+        else:
+            self.text = Text(text, type_=TextType.PLAINTEXT, verbatim=False)
+
+    def _resolve(self) -> Dict[str, any]:
+        header = self._attributes()
+        header["text"] = self.text._resolve()
+        return header
+
+
 class ImageBlock(Block):
     """
     A simple image block, designed to make those cat photos really pop.
@@ -134,86 +235,88 @@ class ImageBlock(Block):
         return image
 
 
-class ActionsBlock(Block):
+class InputBlock(Block):
     """
-    A block that is used to hold interactive elements.
+    A block that collects information from users - it can hold a plain-text
+    input element, a checkbox element, a radio button element, a select
+    menu element, a multi-select menu element, or a datepicker.
     """
 
     def __init__(
-        self, elements: Optional[List[Element]] = None, block_id: Optional[str] = None
+        self,
+        label: TextLike,
+        element: Element,
+        dispatch_action: Optional[DispatchActionConfiguration] = None,
+        block_id: Optional[str] = None,
+        hint: Optional[TextLike] = None,
+        optional: bool = False,
     ):
-        super().__init__(type_=BlockType.ACTIONS, block_id=block_id)
-        if isinstance(elements, Element):
-            self.elements = [
-                elements,
+        super().__init__(type_=BlockType.INPUT, block_id=block_id)
+        self.label = Text.to_text(
+            label, force_plaintext=True, max_length=2000, allow_none=False
+        )
+        if not isinstance(element, ALLOWED_INPUT_ELEMENTS):
+            raise InvalidUsageError("")
+        self.element = element
+        self.dispatch_action = dispatch_action
+        self.hint = Text.to_text(
+            hint, force_plaintext=True, max_length=2000, allow_none=True
+        )
+        self.optional = optional
+
+    def _resolve(self) -> Dict[str, Any]:
+        input_block = self._attributes()
+        input_block["label"] = self.label._resolve()
+        input_block["element"] = self.element._resolve()
+        if self.dispatch_action:
+            input_block["dispatch_action"] = self.dispatch_action._resolve()
+        if self.hint:
+            input_block["hint"] = self.hint._resolve()
+        if self.optional:
+            input_block["optional"] = self.optional
+        return input_block
+
+
+class SectionBlock(Block):
+    """
+    A section is one of the most flexible blocks available -
+    it can be used as a simple text block, in combination with text fields,
+    or side-by-side with any of the available block elements.
+    """
+
+    def __init__(
+        self,
+        text: Optional[TextLike] = None,
+        block_id: Optional[str] = None,
+        fields: Optional[Union[TextLike, List[TextLike]]] = None,
+        accessory: Optional[Element] = None,
+    ):
+        super().__init__(type_=BlockType.SECTION, block_id=block_id)
+        if not text and not fields:
+            raise InvalidUsageError(
+                "Must supply either `text` or `fields` or `both` to SectionBlock."
+            )
+        self.text = Text.to_text(text, max_length=3000, allow_none=True)
+        self.fields = coerce_to_list(
+            [
+                Text.to_text(field, max_length=2000, allow_none=False)
+                for field in coerce_to_list(fields, class_=(str, Text), allow_none=True)
             ]
-        elif isinstance(elements, list) and all(
-            [isinstance(el, Element) for el in elements]
-        ):
-            self.elements = elements
+            if fields
+            else None,
+            class_=Text,
+            allow_none=True,
+            max_size=10,
+        )
 
-    def _resolve(self):
-        actions = self._attributes()
-        actions["elements"] = [element._resolve() for element in self.elements]
-        return actions
+        self.accessory = accessory
 
-
-class ContextBlock(Block):
-    """
-    Displays message context, which can include both images and text.
-    """
-
-    def __init__(
-        self, elements: Optional[List[Element]] = None, block_id: Optional[str] = None
-    ):
-        super().__init__(type_=BlockType.CONTEXT, block_id=block_id)
-        self.elements = []
-        for element in elements:
-            if element.type == ElementType.TEXT or element.type == ElementType.IMAGE:
-                self.elements.append(element)
-            else:
-                raise InvalidUsageError(
-                    "Context blocks can only hold image and text elements"
-                )
-        if len(self.elements) > 10:
-            raise InvalidUsageError("Context blocks can hold a maximum of ten elements")
-
-    def _resolve(self) -> Dict[str, any]:
-        context = self._attributes()
-        context["elements"] = [element._resolve() for element in self.elements]
-        return context
-
-
-class FileBlock(Block):
-    """
-    Displays a remote file.
-    """
-
-    def __init__(self, external_id: str, source: str, block_id: Optional[str]):
-        super().__init__(type_=BlockType.FILE, block_id=block_id)
-        self.external_id = external_id
-        self.source = source
-
-    def _resolve(self) -> Dict[str, any]:
-        file = self._attributes()
-        file["external_id"] = self.external_id
-        file["source"] = self.source
-        return file
-
-
-class HeaderBlock(Block):
-    """
-    A header is a plain-text block that displays in a larger, bold font.
-    """
-
-    def __init__(self, text: Union[str, Text], block_id: Optional[str] = None):
-        super().__init__(type_=BlockType.HEADER, block_id=block_id)
-        if type(text) is Text:
-            self.text = text
-        else:
-            self.text = Text(text, type_=TextType.PLAINTEXT, verbatim=False)
-
-    def _resolve(self) -> Dict[str, any]:
-        header = self._attributes()
-        header["text"] = self.text._resolve()
-        return header
+    def _resolve(self) -> Dict[str, Any]:
+        section = self._attributes()
+        if self.text:
+            section["text"] = self.text._resolve()
+        if self.fields:
+            section["fields"] = [field._resolve() for field in self.fields]
+        if self.accessory:
+            section["accessory"] = self.accessory._resolve()
+        return section
